from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional
from helmet.utils.constants import FEATURE_ATTRIBUTION, ALTERNATIVES, CONTRASTIVE, CERTAINTY

@dataclass 
class Attribution:
   """Attribution dataclass"""
   attribution: list[float] = field(default_factory=list)

@dataclass
class Explanation:
    """Generic explanation dataclass"""
    explanation_method: str

@dataclass
class FeatureAttributionExplainer(Explanation):
    """Saliency Explanation dataclass"""
    input_attributions: Attribution | list[Attribution]
    def __init__(self, input_attributions: Attribution | list[Attribution]):
        self.input_attributions = input_attributions
        super().__init__(FEATURE_ATTRIBUTION)
    
@dataclass
class ContrastiveExplanation(Explanation):
    """Contrastive Explanation dataclass"""
    contrastive_input: str
    attributions: Attribution
    def __init__(self, contrastive_input: str, attributions: Attribution):
        self.contrastive_input = contrastive_input
        self.attributions = attributions
        super().__init__(CONTRASTIVE)

@dataclass
class AlternativesExplanation(Explanation):
    """Alternatives Explanation dataclass"""
    output_alternatives: list[list[dict[str, float]]]
    def __init__(self, output_alternatives: list[list[dict[str, float]]]):
        self.output_alternatives = output_alternatives
        super().__init__(ALTERNATIVES)

@dataclass
class CertaintyExplanation(Explanation):
    """Certainty Explanation dataclass"""
    certainties: list[float]
    def __init__(self, certainties: list[float]):
        self.certainties = certainties
        super().__init__(CERTAINTY)

explanation_name_to_class = {
    ALTERNATIVES: AlternativesExplanation,
    CONTRASTIVE: ContrastiveExplanation,
    FEATURE_ATTRIBUTION: FeatureAttributionExplainer,
    CERTAINTY: CertaintyExplanation
}

@dataclass  
class Input:
    """Generic Prompt Class"""
    prompt: str
    input_tokens: list[str]
    def dict(self) -> dict:
        return {
            "prompt": self.prompt,
            "input_tokens": self.input_tokens
        }

@dataclass
class ContextInput(Input):
    """Prompt & Context"""
    context: str
    context_tokens: list[str]
    def dict(self) -> dict:
        d = super().dict()
        d["context"] = self.context
        d["context_tokens"] = self.context_tokens
        return d

@dataclass
class Output:
    """Generic output dataclass"""
    output_str: str
    tokens: list[str]
    def dict(self) -> dict:
        return {
            "output_str": self.output_str,
            "tokens": self.tokens
        }

@dataclass
class Run:
    """Generic run dataclass"""
    project_id: str
    date: datetime
    model_checkpoint: str
    tokenizer: str
    model_type: str
    input: Input
    output: Output
    explanations: list[Explanation]

    _id: Optional[str] = None
    groundtruth: Optional[str] = None
    execution_time_in_sec: Optional[float] = None
    
    def dict(self) -> dict:
        d = {
            "date": self.date,
            "model_checkpoint": self.model_checkpoint,
            "tokenizer": self.tokenizer,
            "model_type": self.model_type,
            "input": self.input.dict(),
            "output": self.output.dict(),
            "explanations": self.explanations,
            "project_id": self.project_id,
            "execution_time_in_sec": self.execution_time_in_sec
        }
        if self._id is not None:
            d["_id"] = self._id
        if self.groundtruth is not None:
            d["groundtruth"] = self.groundtruth
        return d
    



import time
from operator import attrgetter
from typing import Tuple
import numpy as np

import torch
import transformers

from helmet.explainers.gradients import analyze_token, input_x_gradient
from helmet.model.base_lm import Base_LM
from helmet.utils.types import *


class DEC_LM(Base_LM):
    def __init__(self, 
                 model_checkpoint: str, 
                 model: transformers.AutoModelForCausalLM, 
                 tokenizer: transformers.PreTrainedTokenizer, 
                 url: str, 
                 project_id: str, 
                 model_config: dict = {}, 
                 device="cpu"):
        self.model_type = "dec"
        self.model_config = model_config

        try:
            assert "embeddings" in model_config, AssertionError("embeddings must be specified in model_config")
            retriever = attrgetter(model_config["embeddings"])
            embeddings = retriever(model)
            assert embeddings is not None, AssertionError(f"embeddings {model_config['embeddings']} not found in model")
        except Exception as e:
            print(e)
            raise KeyError("embeddings must be specified in model_config")

        super().__init__(model_checkpoint, model, tokenizer, self.model_type, url, project_id, embeddings, device)
    
    def forward(self, inputs, generation_args, **kwargs) -> Tuple[list, CertaintyExplanation]:
        input_len = len(inputs["input_ids"][0])

        generated_outputs = self.model.generate(
            **inputs, 
            return_dict_in_generate=True,
            output_scores=True, # this gets the scores
            **generation_args,
            **kwargs
        )
        
        transition_scores = self.model.compute_transition_scores(generated_outputs.sequences, generated_outputs.scores, normalize_logits=True)
        gen_sequences = generated_outputs.sequences[:, input_len:]
        certainties = [float(np.exp(score.cpu().numpy())) for score in transition_scores[0]]

        outs = gen_sequences[0].detach().cpu().numpy() 

        return outs, CertaintyExplanation(certainties)
    
    def predict(self, prompt, generation_args, groundtruth=None, *args, **kwargs):
        start = time.time()
        input = self._encode_text(prompt)
        input_str = self.token_ids_to_string(input["input_ids"][0])
        
        print("Processed input:", input_str)
        
        output_token_ids, certainties = self.forward(input, generation_args)
        output_str: str = self.token_ids_to_string(output_token_ids)

        print("Output:", output_str)
        
        end = time.time()
        execution_time = end - start

        formatted_run = self._format_run(input_str, output_str, [certainties], execution_time, groundtruth=groundtruth)

        id = self.update_run(formatted_run)

        return output_str, id

    def feature_attribution(self, id: str, **kwargs) -> FeatureAttributionExplainer:
        run: Run = self.get_run(id)
        input = self._encode_text(run.input.prompt) #on cuda
        output_token_ids = self.tokenizer.convert_tokens_to_ids(run.output.tokens) #on cpu
        output_token_ids = torch.tensor(output_token_ids).to(self.device)
        input_ids = input["input_ids"][0]
        attention_mask = input["attention_mask"]

        merged = torch.cat((input_ids, output_token_ids), 0)

        start_index = len(input_ids)
        total_length = len(merged)

        result = []
        for idx in range(start_index, total_length):
            curr_input_ids = merged[:idx]
            output_id = merged[idx]
            base_saliency_matrix, base_embd_matrix = analyze_token(self, curr_input_ids, attention_mask, correct=output_id)
            gradients = input_x_gradient(base_saliency_matrix, base_embd_matrix, normalize=True)
            result.append(gradients)
            print("finished token", idx, "of", total_length)

        explanation = FeatureAttributionExplainer(input_attributions=result)
        run.explanations.append(explanation)

        id = self.update_run(run)

        return explanation
    
    def contrastive_explainer(self, id: str, alternative_str: str, **kwargs) -> ContrastiveExplanation:
        run: Run = self.get_run(id)
        input = self._tokenize(run.input.prompt)
        alternative_output = self._encode_text(alternative_str)
        output_token_ids = self.tokenizer.convert_tokens_to_ids(run.output.tokens)

        input_ids = self.to_device(input["input_ids"][0])
        attention_mask = self.to_device(input["attention_mask"])
        
        output_id = output_token_ids[0]

        alternative_id = alternative_output["input_ids"][0]
        if len(alternative_id) > 1:
            print("Warning: alternative output has more than one token, using the first one")
            alternative_id = alternative_id[0]
        saliency_matrix, base_embd_matrix = analyze_token(self, input_ids, attention_mask, correct=output_id, foil=alternative_id)
        gradients = input_x_gradient(saliency_matrix, base_embd_matrix, normalize=True)

        alternative_output_str = self.tokenizer.decode(alternative_id, skip_special_tokens=True)
        
        explanation = ContrastiveExplanation(contrastive_input=alternative_output_str, attributions=gradients)
        run.explanations.append(explanation)

        id = self.update_run(run)

        return explanation
        




import torch
import transformers
from transformers import AutoTokenizer

from helmet.model import DEC_LM
from helmet.updater import get_or_create_project

url = "http://localhost:4000"

# Mapping from model_type to model class
model_type_to_class = {
    "enc": transformers.AutoModelForSequenceClassification,
    "dec": transformers.AutoModelForCausalLM,
    "enc-dec": transformers.AutoModelForSeq2SeqLM,
}

model_type_to_model_wrapper = {
    # "enc": ENC_LM,
    # "enc-dec": ENC_DEC_LM,
    "dec": DEC_LM,
}


def from_pretrained(model_checkpoint: str, model_type: str, embeddings_path:str, project_id:str,                     device:str, platform_url:str = url, model_config: dict={}):
    """
    Load a model from a checkpoint and return the model object
    Args:
        model_checkpoint (str): The model checkpoint to load
        model_type (str): The type of model to load (enc, dec, enc-dec)
        embeddings_path (str): The embeddings pointer
        project_id (str): The project id to send updates to
        device (str): The device to load the model on (cpu or cuda)
        platform_url (str): The url to send updates to (default is http://localhost:4000)
        model_config (dict): The model configuration
    """
    # First checks
    assert device in ["cpu", "cuda"], AssertionError("device must be either 'cpu' or 'cuda'")
    assert model_type in ["enc", "dec", "enc-dec"], AssertionError("model_type must be either 'enc', 'dec', or 'enc-dec'")
    if model_type in ["enc", "enc-dec"]:
        raise NotImplementedError("model_type 'enc' and 'enc-dec' not implemented yet")
    
    if device == "cuda":
        assert torch.cuda.is_available(), AssertionError("cuda is not available")
        torch.device(device)

    
    print("updates will be sent to", platform_url)
    print("setting up model with config, ", model_config)

    # Getting the Hugginface class
    model_cls = model_type_to_class[model_type]

    # Quantization of the model
    if device == "cuda":
        from transformers import BitsAndBytesConfig
        quantization_config = BitsAndBytesConfig(llm_int8_enable_fp32_cpu_offload=True, load_in_8bit=True)
        hfModel = model_cls.from_pretrained(model_checkpoint, trust_remote_code=True, config=quantization_config, 
                                            device_map="auto", torch_dtype = torch.float16 if device == "cuda" else torch.float32, **model_config)
        hfModel = hfModel.to(device)
        
    else:
        hfModel = model_cls.from_pretrained(model_checkpoint, trust_remote_code=True, **model_config)

    hfTokenizer = AutoTokenizer.from_pretrained(model_checkpoint, **model_config)
    modelHelper = model_type_to_model_wrapper[model_type]

    model_setup = {
        "embeddings": embeddings_path
    }

    model = modelHelper(model_checkpoint, hfModel, hfTokenizer, platform_url, project_id, model_setup, device=device)
    return model


__all__ = ["from_pretrained", "get_or_create_project"]